/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * Respeed test.
 *
 * Function: demo_respeed
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include "../gdamchannel.h"

typedef struct _DemoInfo {
	GdamChannel* channel;
	GtkWindow* top_level;
	GtkLabel* filename;
	GtkFileSelection* file_selection;
	guint mixer_id;
	guint respeed_id;
	gboolean making_request;
} DemoInfo;

static void chose_file(GtkButton* button, DemoInfo* info)
{
	char* filename;
	g_return_if_fail(info->file_selection != NULL);
	filename = gtk_file_selection_get_filename(info->file_selection);
	gtk_label_set_text(info->filename, filename);
}

static void filesel_button_clicked(GtkButton* button, DemoInfo* info)
{
	if (info->file_selection) {
		GtkObject* to_kill = GTK_OBJECT(info->file_selection);
		info->file_selection = NULL;
		gtk_object_destroy(to_kill);
	} else {
		GtkWidget* filesel = gtk_file_selection_new("pick your mp3");
		GtkFileSelection* file_select = (GtkFileSelection*)filesel;
		info->file_selection = file_select;
		gtk_signal_connect(GTK_OBJECT(file_select->ok_button),
		                   "clicked",
		                   (GtkSignalFunc) chose_file,
				   info);
		/*gtk_signal_connect_object(
		                   GTK_OBJECT(file_select->cancel_button),
		                   "clicked",
				   (GtkSignalFunc) gtk_widget_destroy,
				   GTK_OBJECT(file_select));*/
		gtk_widget_show(GTK_WIDGET(file_select));
	}
}

static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	int mp3_id;
	if (info->respeed_id != 0 || info->making_request)
		return;

	info->making_request = TRUE;
	gtk_label_get(info->filename, &filename);
	gdam_debug("create respeed %s", filename);
	mp3_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamSeekableMp3",
				1,
				"filename",
				GDAM_ARG_TYPE_STRING,
				filename);
	g_assert(mp3_id != 0);
	
	info->respeed_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamRespeed",
				1,
				"filtered",
				GDAM_ARG_TYPE_INT,
				mp3_id);
	g_assert(info->respeed_id != 0);
	gdam_debug("gdam_channel_mix_add being called...");
	if (!gdam_channel_mix_add(info->channel,
	                          info->mixer_id,
				  info->respeed_id,
				  -1))
		g_warning("respeed_add failed");
	info->making_request = FALSE;
}

static void stop_callback(GtkButton* button, DemoInfo* info)
{
	if (info->respeed_id == 0 || info->making_request)
		return;
	if (!gdam_channel_detach_source(info->channel, info->respeed_id))
		g_warning("detach source failed");
	info->respeed_id = 0;
}


/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}

static void respeed_changed(GtkAdjustment* adjustment, DemoInfo* info)
{
	if (info->respeed_id == 0)
		return;
	gdam_channel_source_set_arg(info->channel,
	                            info->respeed_id,
	                            "rate",
				    GDAM_ARG_TYPE_DOUBLE,
			            adjustment->value);
}

#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void demo_respeed(GdamChannel*           channel,
                  guint                  mixer_id)
{
	DemoInfo* info = g_new(DemoInfo, 1);
	GtkWidget* hbox;
	GtkWidget* vbox;
	GtkAdjustment* adjustment;
	GtkWidget* scroll;

	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->file_selection = NULL;
	info->making_request = FALSE;
	info->mixer_id = mixer_id;
	info->respeed_id = 0;

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                  generic_button("File...", filesel_button_clicked));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));


	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Stop", stop_callback));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));


	adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
					1.0, 0.0, 3.0,	/* default, min, max */
					0.01, 0.1, 1.0));
	gtk_signal_connect(GTK_OBJECT(adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(respeed_changed),
			   info);
	scroll = gtk_hscrollbar_new(adjustment);
	gtk_range_set_update_policy(GTK_RANGE(scroll), GTK_UPDATE_CONTINUOUS);
	gtk_widget_show(scroll);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(scroll));

	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
